/*
 * The contents of this file are subject to the terms 
 * of the Common Development and Distribution License 
 * (the "License").  You may not use this file except 
 * in compliance with the License.
 * 
 * You can obtain a copy of the license at 
 * glassfish/bootstrap/legal/CDDLv1.0.txt or 
 * https://glassfish.dev.java.net/public/CDDLv1.0.html. 
 * See the License for the specific language governing 
 * permissions and limitations under the License.
 * 
 * When distributing Covered Code, include this CDDL 
 * HEADER in each file and include the License file at 
 * glassfish/bootstrap/legal/CDDLv1.0.txt.  If applicable, 
 * add the following below this CDDL HEADER, with the 
 * fields enclosed by brackets "[]" replaced with your 
 * own identifying information: Portions Copyright [yyyy] 
 * [name of copyright owner]
 */
// Copyright (c) 1998, 2006, Oracle. All rights reserved.  
package oracle.toplink.essentials.internal.security;

import java.security.*;
import java.lang.reflect.*;

/**
 * INTERNAL:
 * Privileged Access Controller is used to control all TopLink calls that need privileged
 * security access.  It wraps those calls in such a way that it is possible to disable the use
 * of doPrivileged for performance reasons.
 * In general, calls will be made to doPrivileged if and only if a java security manager is
 * enabled.
 * For users that wish to use a security manager and disable the use of doPrivileged, users can
 * set one of two system flags (through the java -Dxxxxx option):
 *
 * oracle.j2ee.toplink.security.usedoprivileged=false
 * oracle.j2ee.security.usedoprivileged=false
 */
public class PrivilegedAccessController {
    private static boolean shouldUsePrivilegedAccess = false;
    private static boolean shouldSecurityManagerBeChecked = true;

    /**
     * Finding a field within a class potentially has to navigate through it's superclasses to eventually
     * find the field.  This method is called by the public getDeclaredField() method and does a recursive
     * search for the named field in the given classes or it's superclasses.
     */
    private static Field findDeclaredField(Class javaClass, String fieldName) throws NoSuchFieldException {
        try {
            return javaClass.getDeclaredField(fieldName);
        } catch (NoSuchFieldException ex) {
            Class superclass = javaClass.getSuperclass();
            if (superclass == null) {
                throw ex;
            } else {
                return findDeclaredField(superclass, fieldName);
            }
        }
    }

    /**
     * Finding a method within a class potentially has to navigate through it's superclasses to eventually
     * find the method.  This method is called by the public getDeclaredMethod() method and does a recursive
     * search for the named method in the given classes or it's superclasses.
     */
    private static Method findMethod(Class javaClass, String methodName, Class[] methodParameterTypes) throws NoSuchMethodException {
        try {
            return javaClass.getDeclaredMethod(methodName, methodParameterTypes);
        } catch (NoSuchMethodException ex) {
            Class superclass = javaClass.getSuperclass();
            if (superclass == null) {
                throw ex;
            } else {
                try{
                    return findMethod(superclass, methodName, methodParameterTypes);
                }catch (NoSuchMethodException lastEx){
                    throw ex;
                }
            }
        }
    }

    /**
     * Execute a java Class.forName().  Wrap the call in a doPrivileged block if necessary.
     * @param className
     */
    public static Class getClassForName(final String className) throws ClassNotFoundException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Class)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws ClassNotFoundException {
                            return Class.forName(className);
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (ClassNotFoundException)exception.getException();
            }
        } else {
            return Class.forName(className);
        }
    }

    /**
     * Execute a java Class.forName() wrap the call in a doPrivileged block if necessary.
     * @param className
     * @param initialize
     * @param loader
     * @throws java.lang.ClassNotFoundException
     */
    public static Class getClassForName(final String className, final boolean initialize, final ClassLoader loader) throws ClassNotFoundException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Class)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws ClassNotFoundException {
                            return Class.forName(className, initialize, loader);
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (ClassNotFoundException)exception.getException();
            }
        } else {
            return Class.forName(className, initialize, loader);
        }
    }

    /**
     * Gets the class loader for a given class.  Wraps the call in a privileged block if necessary
     */
    public static ClassLoader getClassLoaderForClass(final Class clazz) {
        if (shouldUsePrivilegedAccess()) {
            return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return clazz.getClassLoader();
                    }
                });
        } else {
            return clazz.getClassLoader();
        }
    }

    /**
     * Get the public constructor for the given class and given arguments and wrap it in doPrivileged if
     * necessary.  The shouldSetAccessible parameter allows the the setAccessible API to be called as well.
     * This option was added to avoid making multiple doPrivileged calls within InstantiationPolicy.
     * @param javaClass The class to get the Constructor for
     * @param args An array of classes representing the argument types of the constructor
     * @param shouldSetAccessible whether or not to call the setAccessible API
     * @throws java.lang.NoSuchMethodException
     */
    public static Constructor getConstructorFor(final Class javaClass, final Class[] args, final boolean shouldSetAccessible) throws NoSuchMethodException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Constructor)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws NoSuchMethodException {
                            Constructor result = javaClass.getConstructor(args);
                            if (shouldSetAccessible) {
                                result.setAccessible(true);
                            }
                            return result;
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (NoSuchMethodException)exception.getException();
            }
        } else {
            Constructor result = javaClass.getConstructor(args);
            if (shouldSetAccessible) {
                result.setAccessible(true);
            }
            return result;
        }
    }

    /**
     *  Get the context ClassLoader for a thread.  Wrap the call in a doPrivileged block if necessary.
     */
    public static ClassLoader getContextClassLoader(final Thread thread) {
        if (shouldUsePrivilegedAccess()) {
            return (ClassLoader)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return thread.getContextClassLoader();
                    }
                });
        } else {
            return thread.getContextClassLoader();
        }
    }

    /**
     * Get the constructor for the given class and given arguments (regardless of whether it is public
     * or private))and wrap it in doPrivileged if necessary.  The shouldSetAccessible parameter allows
     * the the setAccessible API to be called as well. This option was added to avoid making multiple
     * doPrivileged calls within InstantiationPolicy.
     * @param javaClass The class to get the Constructor for
     * @param args An array of classes representing the argument types of the constructor
     * @param shouldSetAccessible whether or not to call the setAccessible API
     * @throws java.lang.NoSuchMethodException
     */
    public static Constructor getDeclaredConstructorFor(final Class javaClass, final Class[] args, final boolean shouldSetAccessible) throws NoSuchMethodException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Constructor)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws NoSuchMethodException {
                            Constructor result = javaClass.getDeclaredConstructor(args);
                            if (shouldSetAccessible) {
                                result.setAccessible(true);
                            }
                            return result;
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (NoSuchMethodException)exception.getException();
            }
        } else {
            Constructor result = javaClass.getDeclaredConstructor(args);
            if (shouldSetAccessible) {
                result.setAccessible(true);
            }
            return result;
        }
    }

    /**
     * Get a field in a class or its superclasses and wrap the call in doPrivileged if necessary.
     * The shouldSetAccessible parameter allows the the setAccessible API to be called as well.
     * This option was added to avoid making multiple doPrivileged calls within InstanceVariableAttributeAccessor.
     * @param javaClass The class to get the field from
     * @param fieldName The name of the field
     * @param shouldSetAccessible whether or not to call the setAccessible API
     * @throws java.lang.NoSuchFieldException
     */
    public static Field getField(final Class javaClass, final String fieldName, final boolean shouldSetAccessible) throws NoSuchFieldException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Field)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws NoSuchFieldException {
                            Field field = (Field)findDeclaredField(javaClass, fieldName);
                            if (shouldSetAccessible) {
                                field.setAccessible(true);
                            }
                            return field;
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (NoSuchFieldException)exception.getException();
            }
        } else {
            Field field = (Field)findDeclaredField(javaClass, fieldName);
            if (shouldSetAccessible) {
                field.setAccessible(true);
            }
            return field;
        }
    }

    /**
     * Get a field actually declared in a class and wrap the call in doPrivileged if necessary.
     * The shouldSetAccessible parameter allows the the setAccessible API to be called as well.
     * This option was added to avoid making multiple doPrivileged calls within InstanceVariableAttributeAccessor.
     * @param javaClass The class to get the field from
     * @param fieldName The name of the field
     * @param shouldSetAccessible whether or not to call the setAccessible API
     * @throws java.lang.NoSuchFieldException
     */
    public static Field getDeclaredField(final Class javaClass, final String fieldName, final boolean shouldSetAccessible) throws NoSuchFieldException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Field)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                    public Object run() throws NoSuchFieldException {
                        Field field = javaClass.getDeclaredField(fieldName);
                        if (shouldSetAccessible) {
                            field.setAccessible(true);
                        }
                        return field;
                    }
                });  
            } catch (PrivilegedActionException exception) {
                throw (NoSuchFieldException)exception.getException();
            }
        } else {
            Field field = javaClass.getDeclaredField(fieldName);
            if (shouldSetAccessible) {
                field.setAccessible(true);
            }
            return field;
        }
    }

    /**
     * Get the list of fields in a class.  Wrap the call in doPrivileged if necessary
     * Excludes inherited fields.
     * @param clazz the class to get the fields from.
     */
    public static Field[] getDeclaredFields(final Class clazz) {
        if (shouldUsePrivilegedAccess()) {
            return (Field[])AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return clazz.getDeclaredFields();
                    }
                });
        } else {
            return clazz.getDeclaredFields();
        }
    }

    /**
     * Return a method on a given class with the given method name and parameter 
     * types. This call will NOT traverse the superclasses. Wrap the call in 
     * doPrivileged if necessary.
     * @param method the class to get the method from
     * @param methodName the name of the method to get
     * @param methodParameters a list of classes representing the classes of the
     *  parameters of the method.
     */
    public static Method getDeclaredMethod(final Class clazz, final String methodName, final Class[] methodParameterTypes) throws NoSuchMethodException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Method)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                    public Object run() throws NoSuchMethodException {
                        return clazz.getDeclaredMethod(methodName, methodParameterTypes);
                    }
                });  
            } catch (PrivilegedActionException exception) {
                throw (NoSuchMethodException)exception.getException();
            }
        } else {
            return clazz.getDeclaredMethod(methodName, methodParameterTypes);
        }
    }
    
    /**
     * Get a method declared in the given class. Wrap the call in doPrivileged 
     * if necessary. This call will traver the superclasses. The 
     * shouldSetAccessible parameter allows the the setAccessible API to be 
     * called as well. This option was added to avoid making multiple 
     * doPrivileged calls within MethodBasedAttributeAccessor.
     * @param javaClass The class to get the method from
     * @param methodName The name of the method to get
     * @param methodParameterTypes A list of classes representing the classes of the parameters of the mthod
     * @param shouldSetAccessible whether or not to call the setAccessible API
     * @throws java.lang.NoSuchMethodException
     */
    public static Method getMethod(final Class javaClass, final String methodName, final Class[] methodParameterTypes, final boolean shouldSetAccessible) throws NoSuchMethodException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return (Method)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws NoSuchMethodException {
                            Method method = (Method)findMethod(javaClass, methodName, methodParameterTypes);
                            if (shouldSetAccessible) {
                                method.setAccessible(true);
                            }
                            return method;
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (NoSuchMethodException)exception.getException();
            }
        } else {
            Method method = findMethod(javaClass, methodName, methodParameterTypes);
            if (shouldSetAccessible) {
                method.setAccessible(true);
            }
            return method;
        }
    }
    
    /**
     * Get the list of methods in a class. Wrap the call in doPrivileged if 
     * necessary. Excludes inherited methods.
     * @param clazz the class to get the methods from.
     */
    public static Method[] getDeclaredMethods(final Class clazz) {
        if (shouldUsePrivilegedAccess()) {
            return (Method[])AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return clazz.getDeclaredMethods();
                    }
                });
        } else {
            return clazz.getDeclaredMethods();
        }
    }
    
    /**
     * Get the list of fields in a class.  Wrap the call in doPrivileged if necessary
     * @param clazz the class to get the fields from.
     */
    public static Field[] getFields(final Class clazz) {
        if (shouldUsePrivilegedAccess()) {
            return (Field[])AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return clazz.getFields();
                    }
                });
        } else {
            return clazz.getFields();
        }
    }
    
    /**
     * Get the return type for a given method. Wrap the call in doPrivileged if necessary.
     * @param field
     */
    public static Class getFieldType(final Field field) {
        if (shouldUsePrivilegedAccess()) {
            return (Class)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return field.getType();
                    }
                });
        } else {
            return field.getType();
        }
    }

    /**
     * Get the line separator character.
     * Previous versions of TopLink always did this in a privileged block so we will continue to do so.
     */
    public static String getLineSeparator() {
        if (shouldUsePrivilegedAccess()) {
            return (String)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return System.getProperty("file.separator");
                    }
                });
        } else {
            return oracle.toplink.essentials.internal.helper.Helper.cr();
        }
    }

    /**
     * Get the list of parameter types for a given method.  Wrap the call in doPrivileged if necessary.
     * @param method The method to get the parameter types of
     */
    public static Class[] getMethodParameterTypes(final Method method) {
        if (shouldUsePrivilegedAccess()) {
            return (Class[])AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return method.getParameterTypes();
                    }
                });
        } else {
            return method.getParameterTypes();
        }
    }

    /**
     * Get the return type for a given method. Wrap the call in doPrivileged if necessary.
     * @param method
     */
    public static Class getMethodReturnType(final Method method) {
        if (shouldUsePrivilegedAccess()) {
            return (Class)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return method.getReturnType();
                    }
                });
        } else {
            return method.getReturnType();
        }
    }
    
    /**
     * Get the list of methods in a class. Wrap the call in doPrivileged if 
     * necessary. This call will traver the superclasses.
     * @param clazz the class to get the methods from.
     */
    public static Method[] getMethods(final Class clazz) {
        if (shouldUsePrivilegedAccess()) {
            return (Method[])AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        return clazz.getMethods();
                    }
                });
        } else {
            return clazz.getMethods();
        }
    }
    
    /**
     * Get the value of the given field in the given object.
     */
    public static Object getValueFromField(final Field field, final Object object) throws IllegalAccessException {
        // PERF: Direct variable access.
        if (shouldUsePrivilegedAccess) {
            try {
                return AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException {
                            return field.get(object);
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (IllegalAccessException)exception.getException();
            }
        } else {
            return field.get(object);
        }
    }

    /**
     * Get the value of the given field in the given object as an int.
     */
    public static int getIntValueFromField(final Field field, final Object object) throws IllegalAccessException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return ((Integer)AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException {
                            return new Integer(field.getInt(object));
                        }
                    })).intValue();
            } catch (PrivilegedActionException exception) {
                throw (IllegalAccessException)exception.getException();
            }
        } else {
            return field.getInt(object);
        }
    }

    /**
     * Construct an object with the given Constructor and the given array of arguments.  Wrap the call in a
     * doPrivileged block if necessary.
     */
    public static Object invokeConstructor(final Constructor constructor, final Object[] args) throws IllegalAccessException, InvocationTargetException, InstantiationException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException, InvocationTargetException, InstantiationException {
                            return constructor.newInstance(args);
                        }
                    });
            } catch (PrivilegedActionException exception) {
                Exception throwableException = exception.getException();
                if (throwableException instanceof IllegalAccessException) {
                    throw (IllegalAccessException)throwableException;
                } else if (throwableException instanceof InstantiationException) {
                    throw (InstantiationException)throwableException;
                } else {
                    throw (InvocationTargetException)throwableException;
                }
            }
        } else {
            return constructor.newInstance(args);
        }
    }

    /**
     * Invoke the givenMethod on a givenObject using the array of parameters given.  Wrap in a doPrivileged block
     * if necessary.
     */
    public static Object invokeMethod(final Method method, final Object object, final Object[] parameters) throws IllegalAccessException, InvocationTargetException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException, InvocationTargetException {
                            // Ensure the method is accessible.
                            if (!method.isAccessible()) {
                                method.setAccessible(true);
                            }
                            
                            return method.invoke(object, parameters);
                        }
                    });
            } catch (PrivilegedActionException exception) {
                Exception throwableException = exception.getException();
                if (throwableException instanceof IllegalAccessException) {
                    throw (IllegalAccessException)throwableException;
                } else {
                    throw (InvocationTargetException)throwableException;
                }
            }
        } else {
            // Ensure the method is accessible.
            if (!method.isAccessible()) {
                method.setAccessible(true);
            }
            
            return method.invoke(object, parameters);
        }
    }

    /**
     * Get a new instance of a class using the default constructor.  Wrap the call in a privileged block
     * if necessary.
     */
    public static Object newInstanceFromClass(final Class clazz) throws IllegalAccessException, InstantiationException {
        if (shouldUsePrivilegedAccess()) {
            try {
                return AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException, InstantiationException {
                            return clazz.newInstance();
                        }
                    });
            } catch (PrivilegedActionException exception) {
                Exception throwableException = exception.getException();
                if (throwableException instanceof IllegalAccessException) {
                    throw (IllegalAccessException)throwableException;
                } else {
                    throw (InstantiationException)throwableException;
                }
            }
        } else {
            return clazz.newInstance();
        }
    }

    /**
     * Set the value of a given field in the given object with the value given.  Wrap the call in a privileged block
     * if necessary.
     */
    public static void setValueInField(final Field field, final Object object, final Object value) throws IllegalAccessException {
        // PERF: Direct variable access.
        if (shouldUsePrivilegedAccess) {
            try {
                AccessController.doPrivileged(new PrivilegedExceptionAction() {
                        public Object run() throws IllegalAccessException {
                            field.set(object, value);
                            return null;
                        }
                    });
            } catch (PrivilegedActionException exception) {
                throw (IllegalAccessException)exception.getException();
            }
        } else {
            field.set(object, value);
        }
    }

    /**
     * This method checks to see if calls should be made to doPrivileged.
     * In general, if a security manager is enabled, it will return true and if one
     * is not enabled, it will return false.
     * It will, however, always return false if either of the following two java properties is
     * set.
     * oracle.j2ee.toplink.security.usedoprivileged=false
     * oracle.j2ee.security.usedoprivileged=false
     * Note: it is not possible to run TopLink using doPrivileged blocks when there is no SecurityManager
     * enabled.
     */
    private static boolean shouldUsePrivilegedAccess() {
        // We will only detect whether to use doPrivileged once.
        if (shouldSecurityManagerBeChecked) {
            shouldSecurityManagerBeChecked = false;

            Boolean privilegedPropertySet = (Boolean)AccessController.doPrivileged(new PrivilegedAction() {
                    public Object run() {
                        boolean propertySet;

                        // check TopLink and OC4j doPrivileged flag.
                        String usePrivileged = System.getProperty("oracle.j2ee.toplink.security.usedoprivileged");
                        String oc4jUsePrivileged = System.getProperty("oracle.j2ee.security.usedoprivileged");
                        propertySet = (((usePrivileged != null) && usePrivileged.equalsIgnoreCase("false")) || ((oc4jUsePrivileged != null) && oc4jUsePrivileged.equalsIgnoreCase("false")));
                        return new Boolean(propertySet);
                    }
                });
            if (privilegedPropertySet.booleanValue()) {
                shouldUsePrivilegedAccess = false;
            } else {
                shouldUsePrivilegedAccess = (System.getSecurityManager() != null);
            }
        }
        return shouldUsePrivilegedAccess;
    }
}
